package com.fzm.common.service.impl;

import cn.fzm.chain.simplesdk.client.ParaChainClient;
import cn.fzm.chain.simplesdk.constant.TxStatusEnum;
import cn.fzm.chain.simplesdk.model.TxResult;
import cn.hutool.core.bean.BeanUtil;
import cn.hutool.json.JSONUtil;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.fzm.common.constant.RedisConstant;
import com.fzm.common.constant.SystemConstant;
import com.fzm.common.entity.Admin;
import com.fzm.common.entity.CommemorateNft;
import com.fzm.common.entity.Nft;
import com.fzm.common.entity.User;
import com.fzm.common.enums.ResultCode;
import com.fzm.common.exception.GlobalException;
import com.fzm.common.mapper.AdminMapper;
import com.fzm.common.mapper.CommemorateNftMapper;
import com.fzm.common.service.*;
import com.fzm.common.utils.JsonUtil;
import com.fzm.common.utils.RedisUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.redisson.api.RLock;
import org.redisson.api.RSemaphore;
import org.redisson.api.RedissonClient;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.util.Date;
import java.util.HashMap;
import java.util.concurrent.TimeUnit;

/**
 * @author tangtuo
 * @date 2021/7/5 15:09
 */
@Slf4j
@Service
public class CommemorateNftServiceImpl extends ServiceImpl<CommemorateNftMapper, CommemorateNft> implements CommemorateNftService {

    @Resource
    NftChainInfoService nftChainInfoService;

    @Resource
    private NftService nftService;

    @Resource
    private ParaChainClient paraChainClient;

    @Value("${chain.para.contract-name}")
    private String contractName;

    @Resource
    private UserService userService;

    @Resource
    private RedissonClient redisson;

    @Resource
    private RedisUtil redisUtil;

    @Override
    public Integer publish(CommemorateNft commemorateNft) {
        User user = userService.getUserByToken();
        String wallet = user.getWallet();
        commemorateNft.setUserId(user.getId()).
                setPublishAddress(wallet);
        // 获取用户的私钥
        String privkey = paraChainClient.walletDumpPrivkey(wallet);
        // 发行nft
        String txHash = paraChainClient.evmPublishNFT1155(contractName, wallet, privkey, commemorateNft.getCount(), true);
        if (StringUtils.isBlank(txHash) || !txHash.contains("-")) {
            throw GlobalException.newException(ResultCode.FAILED, "nft发行失败");
        }
        String[] split = txHash.split("-");
        String hash = split[0];
        long tokenId = Long.parseLong(split[1]);
        // 构建上链信息
        HashMap<String, String> map = new HashMap<>();
        map.put("hash", commemorateNft.getFileHash());
        String tokenInfo = JSONUtil.toJsonStr(map);
        log.info("加密上链数据: {}", tokenInfo);
        String tradeHash = paraChainClient.evmSetTokenInfo(contractName, tokenId, tokenInfo, wallet, privkey, true);
        if (StringUtils.isBlank(tradeHash)) {
            throw GlobalException.newException(ResultCode.PUBLISH_ERROR, "nft发行失败");
        }

        // 确认交易结果
        TxResult txResult = paraChainClient.cycleConfirmTxWithHash(hash, true, 1000);
        if (!TxStatusEnum.SUCCESS.equals(txResult.getStatus())) {
            throw GlobalException.newException(ResultCode.PUBLISH_ERROR, txResult.getErrMsg().getValue());
        }
        String realHash = paraChainClient.getRealTxHashFromGrp(hash);
        commemorateNft.setNftHash(realHash).setTokenId(tokenId).setPublishTime(new Date());
        save(commemorateNft);
        // 使用nft发行量作为Redisson信号量限制领取次数
        RSemaphore semaphore = redisson.getSemaphore(RedisConstant.COMMEMORATE_NFT_PREFIX + commemorateNft.getId());
        semaphore.trySetPermits(commemorateNft.getCount());
        return commemorateNft.getId();
    }

    @Override
    public Boolean receive(Integer id) throws InterruptedException {
        User user = userService.getUserByToken();
        // 加锁,避免一个用户可以并发领取
        RLock lock = redisson.getLock("receive:nft:" + user.getId());
        boolean tryLock = lock.tryLock(30, TimeUnit.SECONDS);
        if (!tryLock){
            throw GlobalException.newException(ResultCode.RECEIVE_ERROR, "操作频繁");
        }
        try {
            CommemorateNft commemorateNft = getById(id);
            if (commemorateNft == null) {
                throw GlobalException.newException(ResultCode.DATA_ERROR, "没找到此纪念版nft详情,请稍后再试");
            }

            // 先判断当前用户有没有领取过此纪念版nft
            if (redisUtil.sIsMember(RedisConstant.COMMEMORATE_NFT_MEMBERS_PREFIX + id, user.getId().toString())) {
                throw GlobalException.newException(ResultCode.FAILED, "此纪念版nft每个用户只能领取一次");
            }
            // 获取此纪念版nft发行数的信号量
            RSemaphore semaphore = redisson.getSemaphore(RedisConstant.COMMEMORATE_NFT_PREFIX + id);
            // 判断纪念版nft有没有被领取完
            boolean acquire = semaphore.tryAcquire();
            // 领取完毕
            if (!acquire) {
                throw GlobalException.newException(ResultCode.FAILED, "抱歉,此纪念版nft已全部被领取完");
            }
            // 领取纪念版nft
            String hash = paraChainClient.evmTransferNFT1155(contractName, commemorateNft.getPublishAddress(), null, user.getWallet(), commemorateNft.getTokenId(), 1, true);
            if (StringUtils.isBlank(hash)) {
                // 领取失败要释放信号量
                semaphore.release();
                throw GlobalException.newException(ResultCode.FAILED, "nft领取失败");
            }
            // 确认转让结果
            TxResult txResult = paraChainClient.cycleConfirmTxWithHash(hash, true, 1000);
            if (!TxStatusEnum.SUCCESS.equals(txResult.getStatus())) {
                // 领取失败要释放信号量
                semaphore.release();
                throw GlobalException.newException(ResultCode.RECEIVE_ERROR, txResult.getErrMsg().getValue());
            }
            String realHash = paraChainClient.getRealTxHashFromGrp(hash);
            Nft nft = new Nft();
            BeanUtil.copyProperties(commemorateNft, nft, true);
            // 获取当前用户领取nft的领取顺序编号
            int availablePermits = commemorateNft.getCount() - semaphore.availablePermits();
            // nft编号为BJIFF11+5为顺序编号 : BJIFF1100123
            String code = "00000" + availablePermits;
            nft.setNftId("BJIFF11" + code.substring(code.length() - 5));
            nft.setTransferHash(realHash);
            nft.setUserId(user.getId());
            nft.setCreateDate(new Date());
            nft.setUpdateDate(new Date());
            nft.setIsArchives(SystemConstant.BOOLEAN_DATA_TRUE);
            nft.setIsGrant(SystemConstant.BOOLEAN_DATA_FALSE);
            nft.setIsCommemorate(SystemConstant.BOOLEAN_DATA_TRUE);
            nftService.save(nft);
            // 领取完毕,把当前用户id加入到纪念版nft的领取记录中
            redisUtil.sAdd(RedisConstant.COMMEMORATE_NFT_MEMBERS_PREFIX + id, user.getId().toString());
            return true;
        }  finally {
              lock.unlock();
        }

    }

    public static void main(String[] args) {
        int a = 12;
        String code = "00000" + a;
        System.out.println("BJIFF11" + code.substring(code.length() - 5));
    }
}
